Module comp_libs
    use omp_lib
    implicit none
   
    
! This module include subroutines for grid, markov chain, and splines
    
    contains   
    
    
Subroutine numderiv(rx,X,Y,dy_dx,direction)
! Direction=1=>It calulcates numerical RIGHT derivative at point X_i using the finite difference formula (Y_(i+1)-Y(i)/X(i+1)-X(i)).
! Direction=2=>It calulcates numerical Left derivative at point X_i using the finite difference formula (Y_(i)-Y(i-1)/X(i)-X(i-1)).
! Direction=3=>It calulcates numerical middle derivative at point X_i using the finite difference formula (Y_(i+1)-Y(i-1)/X(i+1)-X(i-1)).
! Direction=4=>It calculates derivatives based on cubic spline 

 ! The last two values are identical
implicit none
Integer, intent(in) :: rx, direction
Real(8), intent(in) :: X(rx+2), Y(rx+2)
Real(8), intent(out) :: dy_dx(rx+2)
Real(8) :: X_plus(rx+2), dx(rx+2), Y_plus(rx+2), dY(rx+2),aaa,bbb
Real(8) :: X_minus(rx+2), Y_minus(rx+2)
Real(8) :: c1(4,rx+1)
Integer :: I


If (direction==1) then ! right derivative
    X_plus  = (/ X(2:rx+2), X(rx+2) /)
    dX = X_plus - X
    dX(rx+2) = dX(rx+1)
    if (any(dX==0.d0)) print *, "Division by zero in numderiv"
    
    Y_plus  = (/ Y(2:rx+2), Y(rx+2) /)
    dY = Y_plus - Y
    dY(rx+2) = dY(rx+1)
    
    dY_dX = dY/dX 
else if (direction==2) then !left derivative
    x_minus = (/ x(1), x(1:rx+1)/)
    dx = x-x_minus
    dx(1) = dx(2)
    if (any(dX==0.d0)) print *, "Division by zero in numderiv"
    y_minus = (/Y(1), Y(1:rx+1)/)
    dy = y-y_minus
    dy(1) = dy(2)
    
    dy_dx = dy/dx
else if (direction .eq. 3 ) then ! middle derivative
    x_plus(2:rx+1) = x(3:rx+2)
    x_plus(1) = x(2)
    x_plus(rx+2) = x(rx+2)
    x_minus(2:rx+1) = x(1:rx)
    x_minus(1) = x(1)
    x_minus(rx+2) = x(rx+1)
    
    y_plus(2:rx+1) = y(3:rx+2)
    y_plus(1) = y(2)
    y_plus(rx+2) = y(rx+2)
    y_minus(2:rx+1) = y(1:rx)
    y_minus(1) = y(1)
    y_minus(rx+2) = y(rx+1)
    
    dx=x_plus-x_minus
    dy=y_plus-y_minus
    dy_dx=dy/dx
else if (direction .eq. 4) then
    aaa= x(rx+2)-x(rx+1)
    call CubicNkD1(2,rx,x,y,c1)
    dy_dx(1:rx+1) = c1(2,1:rx+1)
    dy_dx(rx+2) = c1(2,rx+1)+2.0d0*c1(3,rx+1)*aaa + 3.0d0*c1(4,rx+1)*aaa ** 2.0d0
end if
    
    


End subroutine numderiv
    
Subroutine gridlookup(r,grid,valplace,iloc)
! Looking for index of grid about a value
! Code from Aubhik's matlab code
! R(input) :  number of inner grids 
! grid(input) : grid we have
! valplace(input) : value we have
! iloc(output) : index number
    Implicit none
    integer, intent(in) :: r
    Real(8), intent(in) :: valplace, grid(r+2)
    integer, intent(out) :: iloc
    integer :: ilow, ihigh, distance, inow
    Real(8) :: valnow

    ilow = 1
    ihigh = r+2
    distance = 2

Do while (distance > 1)

    inow = floor((ilow + ihigh)/2.);
    valnow = grid(inow);


    if (valnow > valplace) then
        ihigh = inow;
    else
        ilow = inow;
    end if

    distance = ihigh - ilow;

end do

iloc = ilow;

end subroutine gridlookup

Subroutine Grid(gridtype,lb,ub,r,gr)
!gridtype (integer, input) : 1:uniform 2:log10 grid 3: double log10 grid
! lb, ub (input) : lower bound and upper bound for grid
!r (input) : number of inner grid points
! gr (output) : grid with dim r+2
Implicit none
Integer, intent(in) :: gridtype, r
Real(8), intent(in) :: lb, ub
Real(8), intent(out) :: gr(r+2)
Real(8) :: w, aaa, bbb, ccc,xx(r+2)
Integer :: I, J, K

If (gridtype==1) then
    gr(1)= lb
    gr(r+2) = ub
    w = (gr(r+2)-gr(1))/(r+1.)
        Do I =2,r+1
            gr(i) = gr(1) + (i-1) * w
        end do
else if (gridtype==2) then
    xx(r+2) = log10(ub-lb+1.0d0)
    xx(1) = log10(lb-lb+1.0d0)
    w = (xx(r+2)-xx(1))/(r+1.0d0)
    
    Do I = 2,r+1
        xx(i) = xx(1) + real(i-1) * w
    end do
    
    Do I=1,r+2
    gr(i) = 10.**(xx(i))+lb - 1.0d0
    end do
     
     
else if (gridtype==3) then
    xx(r+2) = log10(log10(ub-lb+1.0d0)+1.0d0)
    xx(1) = log10(log10(lb-lb+1.0d0)+1.0d0)
    w = (xx(r+2)-xx(1))/(r+1.)
    
   
    Do I = 2,r+1
        xx(i) = xx(1) + (i-1) * w
    end do
 
    Do I=1,r+2
        gr(i) = 10.**(10.**(xx(i))+lb-1.0d0) - 1.0d0
    end do
    
     
else if (gridtype==4) then
    xx(r+2) = log10(log10(log10(ub-lb+1.0d0)+1.0d0)+1.0d0)
    xx(1) = log10(log10(log10(lb-lb+1.0d0)+1.0d0)+1.0d0)
    w = (xx(r+2)-xx(1))/(r+1.)
    
   
    Do I = 2,r+1
        xx(i) = xx(1) + (i-1) * w
    end do
 
    Do I=1,r+2
        gr(i) = 10.**(10.**(10.**(xx(i))+lb-1.0d0)-1.0d0) - 1.0d0
    end do
    
     

     
end if

end subroutine Grid
    
           
subroutine gridroutine(gridtype,Nnk,Npk,lb,ub,gridout)
!gridtype (integer, input) : 1:uniform 2:log10 grid 3: double log10 grid
! Nnk (integer, input) : # of negative grid
! Npk (integer, input) : # of zero and positive grid
! lb, up (real,input) : lower bound and upper bound for grid
! gridout (real,output) : grid with dim Nnk+Npk
    Implicit none
Integer, intent(in) :: gridtype, Nnk, Npk
Real(8), intent(in) :: lb, ub
Real(8), intent(out) :: gridout(Nnk+Npk)
Real(8) :: aaa, ngrid(Nnk+1), pgrid(Npk), tlow, thigh
Integer :: i, j, i1, n1, n2, n3, rnk, rpk
    
tlow=0.0d0
rnk=Nnk+1-2
rpk=Npk-2

If (lb<0.0d0) then
    thigh=-lb
    call Grid(gridtype,tlow,thigh,rnk,ngrid)
    Do i1= 1, Nnk
        gridout(i1) = -ngrid(Nnk+2-i1)
    end do
    thigh=ub
    call Grid(gridtype,tlow,thigh,rpk,pgrid)
    gridout(Nnk+1:Nnk+Npk)=pgrid
else
    tlow=lb
    thigh=ub
    call Grid(gridtype,tlow,thigh,Nnk+Npk-2,gridout)
end if

end subroutine gridroutine

 Subroutine TriDiLU(A,X,B)
 ! subroutine for the solution of tridiagonal matrix (Cubic suppline)
 ! A(input): Ax=B 
 ! X(output) : solution
 ! B(input) : Ax=B
    Implicit none
    Real(8), dimension(:,:), intent(in):: A
    Real(8), dimension(:), intent(in) :: B
    Real(8), dimension(size(B)), intent(out) :: X
    Real(8), dimension(size(A,dim=1)) :: L, U, y
    Real(8) :: aa, uu,ll,a1, bb,xx,f(1:size(a,dim=1)), e(2:size(A,dim=1)), d(1:size(A,dim=1))
    Integer(4) :: N, I


    N = size(A,dim=1)

    Do I = 1, N-1
        f(i) = A(i,i+1)
    end do

    Do I= 2, N
        e(i) = A(i,i-1)
    end do

    Do I = 1,N
        d(i) = A(i,i)
    end do

    U(1) = d(1)

    Do I =2,N
        a1=e(i)
        uu=U(i-1)
        L(i) = a1/uu
        aa=d(i)
        ll=l(i)
        a1=f(i-1)
        U(i) = aa-ll*a1
    end do

    y(1) = b(1)
    Do I = 2,N
        bb= b(i)
        ll = l(i)
        a1= y(i-1)
        y(i)= bb-ll*a1
    end do

    bb= y(N)
    uu= u(N)

    x(N) = bb/uu

    Do I = N-1, 1, -1
        bb = y(i)
        a1 = f(i)
        xx = x(i+1)
        uu = u(i)
        x(i) = (bb-a1*xx)/uu
    end do

end subroutine TriDiLU

Subroutine CubicNkD1(category,r,t,f,C)
! Subroutine for one dimensional cubic splines
! category(input) : 1=natural, 2=not-a-knot 
! r(input) : number of interior nots
! t(input) : points on grids
! f(input) : function's values at t
! C(output) : coefficient matrix
    Implicit none
    Integer, intent(in) :: category,r        !inner knot
    Real(8), intent(in) :: t(r+2), f(r+2)
    Real(8), intent(out) :: C(4,r+1)
    Real(8) :: TT(r,r), s(r+2), FF(r) ,w1, wr, d, c1(r+1), c2(r+1), c3(r+1), c4(r+1), AA(r,r), ss(1,r),dt,ts(r+2),Df(r+2)
    Real(8) :: x(r)
    Character :: N
    Integer :: I, J, K, Id(r)

    Do I = 1,r+1
        ts(i) = t(i+1)-t(i)
        d = ts(i)
        Df(i) = (f(i+1)-f(i))/d
    end do


    
    
    
    w1 = ts(2) - ((ts(1)**2.0d0)/ts(2))
    wr = ts(r) - ((ts(r+1)**2.0d0)/ts(r))


    TT=0.0d0




    Do I= 2, r-1
            TT(I,I-1) = ts(i+1)
            TT(I,I) = 2.0d0*(ts(i)+ts(i+1))
            TT(I,I+1) =  ts(i)
    end do
    
    Do I = 2, r-1
        FF(i) = 3.0d0*(ts(i+1)*Df(i)+ts(i)*Df(i+1))
    end do
    
    If (category==1) then
        TT(1,1)=2*ts(1)+(3./2.)*(ts(2))
        TT(1,2)=ts(1)
        TT(r,r)=(3./2.)*ts(r)+2.*ts(r+1)
        TT(r,r-1)=ts(r+1)
        FF(1)=(3./2.)*ts(2)*Df(1)+3.*ts(1)*Df(2)!+((ts(1)*ts(2)*f(1))/4.)
        FF(r)=3.*ts(r+1)*Df(r)+(3./2.)*ts(r)*Df(r+1)!-((ts(r)*ts(r+1)*f(r+2))/4.)
        Call TriDiLU(TT,x,FF)
        s(2:r+1) = x
        s(1)=(3./2.)*Df(1)-(1./2.)*s(2)!-((ts(1)*f(1))/4.)
        s(r+2)=(3./2.)*Df(r+1)-(1./2.)*s(r+1)!-((ts(r+1)*f(r+2))/4.)
    else if (category==2) then
        TT(1,1) = 2.0d0*(ts(1)+ts(2))-w1
        TT(1,2) = ts(1) + ((ts(1)**2.0d0)/ts(2))
        TT(r,r) = 2.0d0*(ts(r)+ts(r+1))-wr
        TT(r,r-1) = ts(r+1) + ((ts(r+1)**2.0d0)/ts(r))
        FF(1) = 3.0d0*(ts(2)*Df(1)+ts(1)*Df(2))-2.0d0*(ts(2)*Df(1)-((ts(1)**2.0d0)/ts(2))*Df(2))
        FF(r) = 3.0d0*(ts(r+1)*Df(r)+ts(r)*Df(r+1))-2.0d0*(ts(r)*Df(r+1)-((ts(r+1)**2.0d0)/ts(r))*Df(r))
        Call TriDiLU(TT,x,FF)
        s(2:r+1) = x 
        s(1) = 2.0d0*(Df(1)-((ts(1)/ts(2))**2.0d0) * Df(2))-(1.0d0-((ts(1)/ts(2))**2.0d0))*s(2)+((ts(1)/ts(2))**2.0d0) *s(3)
        s(r+2) = 2.0d0* (Df(r+1)-((ts(r+1)/ts(r))**2.0d0) * Df(r))-(1.0d0-((ts(r+1)/ts(r))**2.0d0))*s(r+1)+((ts(r+1)/ts(r))**2.0d0) * s(r)
    end if




    Do I = 1, r+1
        c1(i) = f(i)
        c2(i) = s(i)
        c3(i) = 3.0d0 * (Df(i)/ts(i))-2.0d0*(s(i)/ts(i))-(s(i+1)/ts(i))
        c4(i) = -2.0d0*(DF(i)/(ts(i)**2.0d0))+(s(i)/(ts(i)**2.0d0))+(s(i+1)/(ts(i)**2.0d0))
    end do
    C(1,:)=c1
    C(2,:)=c2
    C(3,:)=c3
    C(4,:)=c4

end subroutine CubicNkD1
 
Subroutine CubicvalD1(x,r,xx,ff,C,f)
! subroutine to approximate a function by using one dimentional cubic spline
! x(input) : value at which we want to apporoximate a function
! r(input) : number of inner grid
! xx(input) : grid 
! ff(input) : function at grid
! C(input) : coefficient of cubic spline
! f(ouput) : approximated value
    Implicit none
    Integer, intent(in) :: r !inner knot
    Real(8), intent(in) :: x, xx(r+2),C(4,r+1), ff(r+2)
    Real(8), intent(out) :: f
    Real(8) :: tx,df1,df2
    Integer :: I, J, K,Ix

    If (x >xx(r+2)) then
        tx = x-xx(r+2)
        df1 = C(2,r+1)+2*C(3,r+1)*(xx(r+2)-xx(r+1))+3*c(4,r+1)*(xx(r+2)-xx(r+1))**2.
        df2 = 2*C(3,r+1)+6*C(4,r+1)*(xx(r+2)-xx(r+1))
        f = ff(r+2)! + df1*tx+(0.5)*df2*tx**2.
    else if (x<xx(1)) then
        !tx = x-xx(1)
        !df1 = C(2,1)
        !df2 = 2*C(3,1)
        f= ff(1)! + df1*tx + (0.5)* df2*tx**2.
    else
        call gridlookup(r,xx,x,Ix)
        tx = x-xx(Ix)
        f = c(1,ix)+ c(2,ix)*tx + c(3,ix)*tx**2.0d0+c(4,ix)*tx**3.
    end if

end subroutine CubicvalD1

Subroutine cubicNkD2(rx,ry,tx,ty,f,C)
! subroutine for cubic splines in two dimension with knot-on-not method
! rx(input) : number of interior knots for x axis
! ry(input) : number of interior knots for y axis
! tx(input) : grid of x axis
! ty(input) : grid of y axis
! f (input) : function's value at (tx,ty)
! C(output) : coefficients of f
    Implicit none
    Integer(4), intent(in) :: rx,ry        !inner knot
    Real(8), intent(in) :: tx(rx+2), ty(ry+2), f(rx+2,ry+2)
    Real(8), intent(out) :: C(4,rx+1,4,ry+1)
    Real(8) :: TTx(rx,rx,ry+2), sx(rx+2,ry+2), FFx(rx,ry+2) ,w1x, wrx, dx, c1x(rx+1,ry+2)
    Real(8) :: CX(4,rx+1,ry+2), AAx(rx,rx), dtx,tsx(rx+2),Dfx(rx+2,ry+2),Dfy(4,rx+1,ry+2)
    Real(8) :: TTy(4,rx+1,ry,ry),  FFy(4,rx+1,ry) ,w1y, wry, dy
    Real(8) :: xx(rx,ry+2),y(ry),yy(4,rx+1,ry),sy(4,rx+1,ry+2),dty,tsy(ry+2)
    Real(8) :: TTTx(rx,rx), x(rx), Fx(rx), TTTy(ry,ry), Fy(ry),c1y(4,rx+1,ry+1),c2y(4,rx+1,ry+1)
    Real(8) :: c2x(rx+1,ry+2), c3x(rx+1,ry+2), c4x(rx+1,ry+2),c3y(4,rx+1,ry+1),c4y(4,rx+1,ry+1)
    Character :: N
    Integer(4) :: I, J, K, I1

    Do I = 1,rx+1
        Do J = 1,ry+2
            tsx(i) = tx(i+1)-tx(i)
            dx = tsx(i)
            Dfx(i,j) = (f(i+1,j)-f(i,j))/dx
        end do
    end do


    w1x = tsx(2) - (tsx(1)**2./tsx(2))
    wrx = tsx(rx) - (tsx(rx+1)**2./tsx(rx))


TTx =0

    Do I = 1, ry+2
        TTx(1,1,I) = 2.*(tsx(1)+tsx(2))-w1x
        TTx(1,2,I) = tsx(1) + tsx(1)**2./tsx(2)
        TTx(rx,rx,I) = 2.*(tsx(rx)+tsx(rx+1))-wrx
        TTx(rx,rx-1,I) = tsx(rx+1) + tsx(rx+1)**2./tsx(rx)
    end do

Do J = 1, ry+2
    Do I= 2, rx-1
            TTx(I,I-1,J) = tsx(i+1)
            TTx(I,I,J) = 2*(tsx(i)+tsx(i+1))
            TTx(I,I+1,J) =  tsx(i)
    end do
end do

Do J=1,ry+2
    FFx(1,J) = 3.*(tsx(2)*Dfx(1,J)+tsx(1)*Dfx(2,J))-2.*(tsx(2)*Dfx(1,J)-(tsx(1)**2./tsx(2))*Dfx(2,J))
    FFx(rx,J) = 3.*(tsx(rx+1)*Dfx(rx,J)+tsx(rx)*Dfx(rx+1,J))-2.*(tsx(rx)*Dfx(rx+1,J)-(tsx(rx+1)**2./tsx(rx))*Dfx(rx,J))
    Do I = 2, rx-1
        FFx(i,J) = 3.*(tsx(i+1)*Dfx(i,J)+tsx(i)*Dfx(i+1,J))
    end do
end do

TTTx=0
Fx=0
x=0

Do I = 1, ry+2
    TTTx = TTx(:,:,I)
    Fx = FFx(:,I)
    Call TriDiLU(TTTx,x,Fx)
    xx(:,I) = x
end do

Do I = 1, ry+2
    x=xx(:,I)
    sx(2:rx+1,I) = x
    sx(1,I) = 2.*(Dfx(1,I)-((tsx(1)/tsx(2))**2.) * Dfx(2,I))-(1-((tsx(1)/tsx(2))**2.))*sx(2,I)+((tsx(1)/tsx(2))**2.) *sx(3,I)
    sx(rx+2,I) = 2.* (Dfx(rx+1,I)-((tsx(rx+1)/tsx(rx))**2.) * Dfx(rx,I)) - (1-((tsx(rx+1)/tsx(rx))**2.))*sx(rx+1,I)+((tsx(rx+1)/tsx(rx))**2.) * sx(rx,I)
end do
Do J = 1, ry+2
    Do I = 1, rx+1
        c1x(i,J) = f(i,J)
        c2x(i,J) = sx(i,J)
        c3x(i,J) = 3. * (Dfx(i,J)/tsx(i))-2.*(sx(i,J)/tsx(i))-(sx(i+1,J)/tsx(i))
        c4x(i,J) = -2.*(DFx(i,J)/tsx(i)**2.)+(sx(i,J)/tsx(i)**2.)+(sx(i+1,J)/tsx(i)**2.)
    end do
end do
    CX(1,:,:)=c1x
    Cx(2,:,:)=c2x
    Cx(3,:,:)=c3x
    Cx(4,:,:)=c4x


Do I = 1,ry+1
    Do J = 1,rx+1
        Do K = 1, 4
            tsy(i) = ty(i+1)-ty(i)
            dy = tsy(i)
            Dfy(K,j,i) = (Cx(K,J,I+1)-Cx(K,J,I))/dy
        end do
    end do
end do

    w1y = tsy(2) - (tsy(1)**2./tsy(2))
    wry = tsy(ry) - (tsy(ry+1)**2./tsy(ry))


TTy =0
Do I = 1, 4
    Do J = 1, rx+1
        TTy(I,J,1,1) = 2.*(tsy(1)+tsy(2))-w1y
        TTy(I,J,1,2) = tsy(1) + tsy(1)**2./tsy(2)
        TTy(I,J,ry,ry) = 2.*(tsy(ry)+tsy(ry+1))-wry
        TTy(I,J,ry,ry-1) = tsy(ry+1) + tsy(ry+1)**2./tsy(ry)
    end do
end do

Do I = 1,4
    Do J = 1, rx+1
        Do K= 2, ry-1
            TTy(I,J,K,K-1) = tsy(K+1)
            TTy(I,J,K,K) = 2.*(tsy(K)+tsy(K+1))
            TTy(I,J,K,K+1) =  tsy(K)
        end do
    end do
end do

Do I = 1,4
    Do J = 1,rx+1
        FFy(I,J,1) = 3.*(tsy(2)*Dfy(I,j,1)+tsy(1)*Dfy(I,J,2))-2.*(tsy(2)*Dfy(I,J,1)-(tsy(1)**2./tsy(2))*Dfy(I,J,2))
        FFy(I,J,ry) = 3.*(tsy(ry+1)*Dfy(I,J,ry)+tsy(ry)*Dfy(I,J,ry+1))-2.*(tsy(ry)*Dfy(I,J,ry+1) -(tsy(ry+1)**2./tsy(ry))*Dfy(I,J,ry))
        Do K = 2, ry-1
            FFy(i,J,K) = 3.*(tsy(K+1)*Dfy(I,J,K)+tsy(K)*Dfy(I,J,K+1))
        end do
    end do
end do
TTTy=0
Fy=0
y=0
Do I = 1,4
    Do J= 1, rx+1
        TTTy = TTy(I,J,:,:)
        Fy = FFy(I,J,:)
        Call TriDiLU(TTTy,y,Fy)
        yy(I,J,:) = y
        y=0
    end do
end do

Do I = 1, 4
    Do J = 1, rx+1
        y=yy(I,J,:)
        sy(I,J,2:ry+1) = y
        sy(I,J,1) = 2.*(Dfy(I,J,1)-((tsy(1)/tsy(2))**2.) * Dfy(I,J,2)) - (1-((tsy(1)/tsy(2))**2.))*sy(I,J,2)+((tsy(1)/tsy(2))**2.) *sy(I,J,3)
        sy(I,J,ry+2) = 2.* (Dfy(I,J,ry+1)-((tsy(ry+1)/tsy(ry))**2.) * Dfy(I,J,ry))-(1-((tsy(ry+1)/tsy(ry))**2.))*sy(I,J,ry+1)+((tsy(ry+1)/tsy(ry))**2.) * sy(I,J,ry)
    end do
end do

Do I = 1, 4
    Do J = 1, rx+1
        Do K = 1, ry+1
            c1y(I,J,K)=cx(I,J,K)
            c2y(I,J,K)=sy(I,J,K)
            c3y(I,J,K)=3. * (Dfy(I,J,K)/tsy(K))-2.*(sy(I,J,K)/tsy(K))-(sy(i,J,K+1)/tsy(K))
            c4y(I,J,K) = -2.*(DFy(I,J,K)/tsy(K)**2.)+(sy(I,J,K)/tsy(K)**2.)+(sy(i,J,K+1)/tsy(K)**2.)
        end do
    end do
end do

Do I = 1,4
    Do J = 1, rx+1
        Do I1 =1,ry+1
            C(I,J,1,I1)=c1y(I,J,I1)
            C(I,J,2,I1)=c2y(I,J,I1)
            C(I,J,3,I1)=c3y(I,J,I1)
            C(I,J,4,I1)=c4y(I,J,I1)
        end do
    end do
end do

end subroutine cubicNkD2

Subroutine CubicvalD2(x,y,rx,ry,xx,yy,ff,C,f)
! subroutine to approximate a function by using two dimensional cubic spline coefficient
! x(input) : a point on x axis 
! y(input) : a point on y axis
! rx(input) : number of interior knots on x axis
! ry(input) : number of interior knots on y axis
! xx(input) : grid on x axis
! yy(input) : grid on y axis
! ff(input) : function's value at girds
! C(input) : two dimensional cubic spline coefficient
! f(output) : approximated funtion value not on the grid points
    Implicit none
    Integer, intent(in) :: rx,ry !inner knot
    Real(8), intent(in) :: x,y, xx(rx+2),yy(ry+2),C(4,rx+1,4,ry+1), ff(rx+2,ry+2)
    Real(8), intent(out) :: f
    Real(8) :: tx, ty,dfx,dfy,dfxx,dfyy,dfxy,fs
    Integer :: I, J, K,ix,iy

    If ((x < xx(rx+2)).and.(y<yy(ry+2))) then
        call gridlookup(rx,xx,x,ix)
        call gridlookup(ry,yy,y,iy)
        tx = x-xx(ix)
        ty = y-yy(iy)
        f = C(1,Ix,1,Iy)+C(1,Ix,2,Iy)*ty+C(1,Ix,3,IY)*ty**2.+C(1,Ix,4,Iy)*ty**3.+&
        C(2,Ix,1,Iy)*tx+C(2,Ix,2,Iy)*tx*ty+C(2,Ix,3,Iy)*tx*ty**2.+C(2,Ix,4,Iy)*tx*ty**3.+&
        C(3,Ix,1,Iy)*tx**2.+C(3,Ix,2,Iy)*tx**2.*ty+C(3,Ix,3,Iy)*tx**2.*ty**2.+C(3,Ix,4,Iy)*tx**2.*ty**3.+&
        C(4,Ix,1,Iy)*tx**3.+C(4,Ix,2,Iy)*tx**3.*ty+C(4,Ix,3,Iy)*tx**3.*ty**2.+C(4,Ix,4,Iy)*tx**3.*ty**3.
    else if ((x>=xx(rx+2)).and.(y<yy(ry+2))) then
        call gridlookup(ry,yy,y,iy)
        ty=y-yy(iy)
        ix=rx+1
        tx=xx(rx+2)-xx(rx+1)
        fs = C(1,Ix,1,Iy)+C(1,Ix,2,Iy)*ty+C(1,Ix,3,IY)*ty**2.+C(1,Ix,4,Iy)*ty**3.+&
        C(2,Ix,1,Iy)*tx+C(2,Ix,2,Iy)*tx*ty+C(2,Ix,3,Iy)*tx*ty**2.+C(2,Ix,4,Iy)*tx*ty**3.+&
        C(3,Ix,1,Iy)*tx**2.+C(3,Ix,2,Iy)*tx**2.*ty+C(3,Ix,3,Iy)*tx**2.*ty**2.+C(3,Ix,4,Iy)*tx**2.*ty**3.+&
        C(4,Ix,1,Iy)*tx**3.+C(4,Ix,2,Iy)*tx**3.*ty+C(4,Ix,3,Iy)*tx**3.*ty**2.+C(4,Ix,4,Iy)*tx**3.*ty**3.

        dfx=C(2,ix,1,iy)+C(2,ix,2,iy)*ty+C(2,ix,3,iy)*ty**2.+C(2,ix,4,iy)*ty**3.+ 2*C(3,ix,1,iy)*tx+2*C(3,ix,2,iy)*tx*ty+&
        2*C(3,ix,3,iy)*tx*ty**2.+2*C(3,ix,4,iy)*tx*ty**3.+3*C(4,ix,1,iy)*tx**2.+3*C(4,ix,2,iy)*tx**2.*ty+3*C(4,ix,3,iy)*tx**2.*ty**2.+3*C(4,ix,4,iy)*tx**2.*ty**3.               
        
        dfxx=2*C(3,ix,1,iy)+2*C(3,ix,2,iy)*ty+2*C(3,ix,3,iy)*ty**2.+2*C(3,ix,4,iy)*ty**3.+6*C(4,ix,1,iy)*tx+6*C(4,ix,2,iy)*tx*ty+6*C(4,ix,3,iy)*tx*ty**2.+6*C(4,ix,4,iy)*tx*ty**3.
          
        f=fs+dfx*(x-xx(rx+2))+(0.5)*(dfxx*(x-xx(rx+2))**2.)
    else if ((x<xx(rx+2)).and.(y>=yy(ry+2))) then
        call gridlookup(rx,xx,x,ix)
        tx = x-xx(ix)
        iy= ry+1
        ty=yy(ry+2)-yy(ry+1)
        fs = C(1,Ix,1,Iy)+C(1,Ix,2,Iy)*ty+C(1,Ix,3,IY)*ty**2.+C(1,Ix,4,Iy)*ty**3.+&
        C(2,Ix,1,Iy)*tx+C(2,Ix,2,Iy)*tx*ty+C(2,Ix,3,Iy)*tx*ty**2.+C(2,Ix,4,Iy)*tx*ty**3.+&
        C(3,Ix,1,Iy)*tx**2.+C(3,Ix,2,Iy)*tx**2.*ty+C(3,Ix,3,Iy)*tx**2.*ty**2.+C(3,Ix,4,Iy)*tx**2.*ty**3.+&
        C(4,Ix,1,Iy)*tx**3.+C(4,Ix,2,Iy)*tx**3.*ty+C(4,Ix,3,Iy)*tx**3.*ty**2.+C(4,Ix,4,Iy)*tx**3.*ty**3.
                   
        dfy=C(1,ix,2,iy)+2*C(1,ix,3,iy)*ty+3*C(1,ix,4,iy)*ty**2.+C(2,ix,2,iy)*tx+2*C(2,ix,3,iy)*tx*ty+3*C(2,ix,4,iy)*tx*ty**2.+&
            C(3,ix,2,iy)*tx**2.+2*C(3,ix,3,iy)*tx**2.*ty+3*C(3,ix,4,iy)*tx**2.*ty**2.+C(4,ix,2,iy)*tx**3.+2*C(4,ix,3,iy)*tx**3.*ty+3*C(4,ix,4,iy)*tx**3.*ty**2.
         
        dfyy=2*C(1,ix,3,iy)+2*C(2,ix,3,iy)*tx+2*C(3,ix,3,iy)*tx**2.+2*C(4,ix,3,iy)*tx**3.+6*C(1,ix,4,iy)*ty+6*C(2,ix,4,iy)*tx*ty+6*C(3,ix,4,iy)*tx**2.*ty+6*C(4,ix,4,iy)*tx**3.*ty
        
        f=fs+dfy*(y-yy(ry+2))+(0.5)*(dfyy*(y-yy(ry+2))**2.)
    else if ((x>=xx(rx+2)).and.(y>=yy(ry+2))) then
        fs=ff(rx+2,ry+2)
        iy= ry+1
        ty=yy(ry+2)-yy(ry+1)
        ix=rx+1
        tx=xx(rx+2)-xx(rx+1)
        dfx=C(2,ix,1,iy)+C(2,ix,2,iy)*ty+C(2,ix,3,iy)*ty**2.+C(2,ix,4,iy)*ty**3.+ 2*C(3,ix,1,iy)*tx+2*C(3,ix,2,iy)*tx*ty+&
             2*C(3,ix,3,iy)*tx*ty**2.+2*C(3,ix,4,iy)*tx*ty**3.+3*C(4,ix,1,iy)*tx**2.+3*C(4,ix,2,iy)*tx**2.*ty+3*C(4,ix,3,iy)*tx**2.*ty**2.+3*C(4,ix,4,iy)*tx**2.*ty**3.               
        
        dfy=C(1,ix,2,iy)+2*C(1,ix,3,iy)*ty+3*C(1,ix,4,iy)*ty**2.+C(2,ix,2,iy)*tx+2*C(2,ix,3,iy)*tx*ty+3*C(2,ix,4,iy)*tx*ty**2.+&
            C(3,ix,2,iy)*tx**2.+2*C(3,ix,3,iy)*tx**2.*ty+3*C(3,ix,4,iy)*tx**2.*ty**2.+C(4,ix,2,iy)*tx**3.+2*C(4,ix,3,iy)*tx**3.*ty+3*C(4,ix,4,iy)*tx**3.*ty**2.
        
        dfxx=2*C(3,ix,1,iy)+2*C(3,ix,2,iy)*ty+2*C(3,ix,3,iy)*ty**2.+2*C(3,ix,4,iy)*ty**3.+6*C(4,ix,1,iy)*tx+6*C(4,ix,2,iy)*tx*ty+6*C(4,ix,3,iy)*tx*ty**2.+6*C(4,ix,4,iy)*tx*ty**3.
        
        dfyy=2*C(1,ix,3,iy)+2*C(2,ix,3,iy)*tx+2*C(3,ix,3,iy)*tx**2.+2*C(4,ix,3,iy)*tx**3.+6*C(1,ix,4,iy)*ty+6*C(2,ix,4,iy)*tx*ty+6*C(3,ix,4,iy)*tx**2.*ty+6*C(4,ix,4,iy)*tx**3.*ty
        
        dfxy=C(2,ix,2,iy)+2*C(2,ix,3,iy)*ty+3*C(2,ix,4,iy)*ty**2.+2*C(3,ix,2,iy)*tx+4*C(3,ix,3,iy)*tx*ty+6*C(3,ix,4,iy)*tx*ty**2.+3*C(4,ix,2,iy)*tx**2.+6*C(4,ix,3,iy)*tx**2.*ty+9*C(4,ix,4,iy)*tx**2.*ty**2.
        
        f=fs+dfx*(x-xx(rx+2))+dfy*(y-yy(ry+2))+0.5*(dfxx*(x-xx(rx+2))**2.+2*dfxy*(x-xx(rx+2))*(y-yy(ry+2))+dfyy*(y-yy(ry+2))**2.)  
    end if

end subroutine CubicvalD2

Subroutine Blintp(x,rx,xx,ff,f)
! Subroutine to approximate one dimentional function with linear interpolation
! x(input) : a point in x grid
! rx(input) : number of interior knots on x axis
! xx(input) : grid on x axis
! ff(input) : function's value on girds
! f(output) : approximated funtion value not on the grid points

implicit none
Integer, intent(in) :: rx
Real(8), intent(in) :: x, xx(rx+2),  ff(rx+2)
Real(8), intent(out) :: f
Real(8) :: wx,wy,aa,bb,fff,cc,dd
Integer :: I, J, K,ix,iy

If (x<=xx(1)) then
   ! aa=ff(2)-ff(1)
   ! bb=xx(2)-xx(1)
   ! cc=aa/bb
   ! dd=(x-xx(1))
    f=ff(1)!+cc*dd
else If (x<xx(rx+2)) then
    aa = x
    call gridlookup(rx,xx,aa,ix)
    aa= xx(ix+1)-x
    bb=xx(ix+1)-xx(ix)
    wx=aa/bb
    f= ff(ix)*wx + ff(ix+1)*(1.-wx)
else if (x>=xx(rx+2)) then
    aa=ff(rx+2)-ff(rx+1)
    bb=xx(rx+2)-xx(rx+1)
    cc=aa/bb
    dd=(x-xx(rx+2))
    f=ff(rx+2)+(cc*dd) !!! Linear extrapolation !!
    !print*, f,x,rx, 'binding'
  
end if

end subroutine Blintp


Subroutine BLintp2(x,y,rx,ry,xx,yy,ff,f)
! Subroutine to approximate two dimentional function with bilinear interpolation
! x(input) : a point in x grid
! y(input) : a point in y grid
! rx(input) : number of interior knots on x axis
! ry(input) : number of interior knots on y axis
! xx(input) : grid on x axis
! yy(input) : grid on y axis
! ff(input) : function's value on girds
! f(output) : approximated funtion value not on the grid points

implicit none
Integer, intent(in) :: rx, ry
Real(8), intent(in) :: x, y, xx(rx+2), yy(ry+2), ff(rx+2,ry+2)
Real(8), intent(out) :: f
Real(8) :: wx,wy,aa,bb,fff
Integer :: I, J, K,ix,iy


If ((x<xx(rx+2)).and.(y<yy(ry+2))) then
    aa=x
    bb=y
    call gridlookup(rx,xx,aa,ix)
    call gridlookup(ry,yy,bb,iy)
    aa=xx(ix+1)-x
    bb=xx(ix+1)-xx(ix)
    wx=aa/bb
    aa=yy(iy+1)-y
    bb=yy(iy+1)-yy(iy)
    wy=aa/bb
    f= ff(ix,iy)*wx*wy+ff(ix+1,iy)*(1.-wx)*wy+ff(ix,iy+1)*wx*(1.-wy)+ff(ix+1,iy+1)*(1.-wx)*(1.-wy)
else if ((x>=xx(rx+2)).and.(y<yy(ry+2))) then
    !wx=(xx(rx+2)-x)/(xx(rx+2)-xx(rx+1))
    call gridlookup(ry,yy,bb,iy)
    aa=yy(iy+1)-y
    bb=yy(iy+1)-yy(iy)
    wy=aa/bb
    !f=ff(rx+1,iy)*wx*wy+ff(rx+2,iy)*(1.-wx)*wy+ff(rx+1,iy+1)*wx*(1.-wy)+ff(rx+2,iy+1)*(1.-wx)*(1.-wy)
    f=ff(rx+2,iy)*wy+ff(rx+2,iy+1)*(1.-wy)
else if ((x<xx(rx+2)).and.(y>=yy(ry+2))) then
    !wy=(yy(ry+2)-y)/(yy(ry+2)-yy(ry+1))
    call gridlookup(rx,xx,aa,ix)
    aa=xx(ix+1)-x
    bb=xx(ix+1)-xx(ix)
    wx=aa/bb
    !f=ff(ix,ry+1)*wx*wy+ff(ix+1,ry+1)*(1.-wx)*wy+ff(ix,ry+2)*wx*(1.-wy)+ff(ix+1,ry+2)*(1.-wx)*(1.-wy)
    f=ff(ix,ry+2)*wx+ff(ix+1,ry+2)*(1.-wx)
else 
    !wx=(xx(rx+2)-x)/(xx(rx+2)-xx(rx+1))
    !wy=(yy(ry+2)-y)/(yy(ry+2)-yy(ry+1))
    !f=ff(rx+1,ry+1)*wx*wy+ff(rx+2,ry+1)*(1.-wx)*wy+ff(rx+1,ry+2)*wx*(1.-wy)+ff(rx+2,ry+2)*(1.-wx)*(1.-wy)
    f=ff(rx+2,ry+2)
end if

end subroutine BLintp2



Subroutine BTintp (x, ix,iy ,oy)
! two dimensional Interpolation on a triangular unstructured grid
!x(input) : point we want to know
!ix(input) : triangle grid points
!iy(input) : function values at the grid points
!oy(output) : interpolated value
Implicit none
Real(8), intent(in) :: x(2),ix(3,2),iy(3)
Real(8), intent(out) :: oy
Real(8) :: L(3),a,b
Integer :: I, J, K

a= (ix(2,2)-ix(3,2))*(x(1)-ix(3,1))+(ix(3,1)-ix(2,1))*(x(2)-ix(3,2))
b= (ix(2,2)-ix(3,2))*(ix(1,1)-ix(3,1))+(ix(3,1)-ix(2,1))*(ix(1,2)-ix(3,2))
L(1)=a/b
a=(ix(3,2)-ix(1,2))*(x(1)-ix(3,1))+(ix(1,1)-ix(3,1))*(x(2)-ix(3,2))
L(2)=a/b
L(3)=1.-L(1)-L(2)

Do I = 1, 3
If ((L(i)<0).or.(L(i)>1)) then
Write(*, '(a)' ,Advance='no')  'Outer Triangle' 
end if
end do

oy = L(1)*iy(1) + L(2)*iy(2) + L(3)*iy(3)
end subroutine BTintp

Subroutine gridsearch(Ng,value,inigrid,endgrid,solgrid,solval)
Implicit none
Integer, intent(in) :: Ng,inigrid,endgrid
Real(8), intent(in) :: value(Ng)
Integer, intent(out) :: solgrid
Real(8), intent(out) :: solval
Real(8) ::temp
Integer :: i1

solgrid=inigrid
solval=value(inigrid)

Do i1= inigrid+1, endgrid
    temp=value(i1)
    If (temp>solval) then
        solval=temp
        solgrid=i1
    end if
end do
end subroutine

Subroutine indandwgt(input,notsk,grid,index,weight)
implicit none

Integer, intent(in) :: notsk
Real(8), intent(in) :: input,grid(notsk+2)
Integer, intent(out) :: index
Real(8), intent(out) :: weight
Integer::klow1


If (input .lt. grid(1)) then
    weight = 1.0d0
    index = 1
else if (input .ge. grid(notsk+2)) then
    weight = 0.0d0
    index = notsk+2-1
else
    call gridlookup(notsk,grid,input,klow1)
    weight=(grid(klow1+1)-input)/(grid(klow1+1)-grid(klow1))
    index=klow1
end if



If ((weight .lt. 0.0d0) .or. (weight .gt. 1.0d0)) then
    print*, 'check the index and program indandwgt, input, weight=', input,weight
    stop
end if
end subroutine indandwgt

Subroutine indandwgt_rst(input,ini,lst,gini,glst,grid,index,weight)
implicit none

Integer, intent(in) :: ini,lst,gini,glst
Real(8), intent(in) :: input,grid(gini:glst)
Integer, intent(out) :: index
Real(8), intent(out) :: weight
Real(8) :: tempgr(lst-ini+1)
Integer::klow1, tempnot

tempnot=lst-ini+1-2
tempgr(1:lst-ini+1)=grid(ini:lst)
call gridlookup(tempnot,tempgr,input,klow1)
!print*, 'kk',klow1, tempgr(klow1:klow1+1), input, ini, lst
!print*, 'kkk', input, grid(ini), grid(lst)
klow1 = klow1 + ini - 1

If (input.lt.grid(ini)) then
    weight = 1.0d0
    index = ini
else if (input .gt. grid(lst)) then
    weight = 0.0d0
    index = lst-1
else
    weight=(grid(klow1+1)-input)/(grid(klow1+1)-grid(klow1))
    index=klow1
end if





If ((weight .lt. 0.0d0) .or. (weight .gt. 1.0d0)) then
    print*, 'check the index and program indandwgt, weight=', weight, ini, lst, gini, glst
    print*, 'input, klow1, index=:', input, klow1, index 
    stop
end if
end subroutine indandwgt_rst


subroutine gen_tauchen (Nz, Mu, rho, Se, m,Pr, z)
! Tauchen algorithm ( Apporoximation of continuous autoregressive process with a Markov Chain )
! This is made by Youngsoo Jang in Jan 2016.
! explanation of variables
! log(z')=mu*(1-Rho)+Rho*log(z)+epsilon, epsilon~N(0,Se^2), y=e^z
! Nz(Input) : the number of grids in terms of z
! Mu(Input) : Long-run mean z
! Rho(Input) : the persistence of Ar(1) process
! Se(Input) : standard deviation of original z (not standard deviation of log(z))!!
! m(input) : max +- std.dev
! Pr(output) : approximated markov chain
! z(output) : grid on Z
implicit none
Real(8), intent(in) :: Mu, rho
Real(8), intent(in) :: Se, m
Integer, intent(in) :: Nz
real(8) :: w, X1, XN,a,b,c,d,e,aaa
Integer:: I, J
Real(8), intent(out) :: pr(Nz,Nz), z(nz)
Real(8)  :: X(Nz),S
Real(8) :: K(nz,nz),K1(nz,nz)
S = sqrt(Se**2./(1.-(rho)**2.))

x1 = -m * S
XN = -X1
w = (XN-X1)/(Nz-1.0d0)

Do I = 1, Nz
    X(I) = x1 + real(I-1)*w
end do

aaa=(1.-rho)*mu
X=X+(aaa/(1.-rho))

Z = exp(X)
d=se
K=0
K1=0
e=0
Do I = 1, Nz
    Do J= 1, Nz-1
        a= x(j)-aaa-rho*x(i)+(w/2.0d0)
        call normal_cdf(a,e,d,c)
        K(i,j)=c
        c=0
    end do
    Do J=2,Nz
        a=x(j)-aaa-rho*x(i)-(w/2.0d0)
        call normal_cdf(a,e,d,c)
        K1(i,j)=c
        c=0
    end do
end do


Do I = 1, Nz
    Pr(i,1) = K(i,1)
end do

Do I = 1,Nz
    do J=2, Nz-1
        Pr(i,j) = K(i,j)-K1(i,j)
    end do
end do

Do I = 1,Nz
    Pr(i,Nz) = 1 - K1(i,Nz)
end do

end subroutine gen_tauchen       


subroutine normal_01_cdf ( x, cdf )
!
!*******************************************************************************
!
!! NORMAL_01_CDF evaluates the Normal 01 CDF.
!
!
!  Reference:
!
!    A G Adams,
!    Areas Under the Normal Curve,
!    Algorithm 39,
!    Computer j.,
!    Volume 12, pages 197-198, 1969.
!
!  Modified:
!
!    10 February 1999
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, real X, the argument of the CDF.
!
!    Output, real CDF, the value of the CDF.
!
  implicit none
!
  real(8), parameter :: a1 = 0.398942280444E+00
  real(8), parameter :: a2 = 0.399903438504E+00
  real(8), parameter :: a3 = 5.75885480458E+00
  real(8), parameter :: a4 = 29.8213557808E+00
  real(8), parameter :: a5 = 2.62433121679E+00
  real(8), parameter :: a6 = 48.6959930692E+00
  real(8), parameter :: a7 = 5.92885724438E+00
  real(8), parameter :: b0 = 0.398942280385E+00
  real(8), parameter :: b1 = 3.8052E-08
  real(8), parameter :: b2 = 1.00000615302E+00
  real(8), parameter :: b3 = 3.98064794E-04
  real(8), parameter :: b4 = 1.98615381364E+00
  real(8), parameter :: b5 = 0.151679116635E+00
  real(8), parameter :: b6 = 5.29330324926E+00
  real(8), parameter :: b7 = 4.8385912808E+00
  real(8), parameter :: b8 = 15.1508972451E+00
  real(8), parameter :: b9 = 0.742380924027E+00
  real(8), parameter :: b10 = 30.789933034E+00
  real(8), parameter :: b11 = 3.99019417011E+00
  real(8),intent(out):: cdf
  real(8) ::  q
  real(8),intent(in) :: x
  real(8) :: y
!
!  |X| <= 1.28.
!
  if ( abs ( x ) <= 1.28 ) then

    y = 0.5E+00 * x**2

    q = 0.5E+00 - abs ( x ) * ( a1 - a2 * y / ( y + a3 - a4 / ( y + a5 &
      + a6 / ( y + a7 ) ) ) )
!
!  1.28 < |X| <= 12.7
!
  else if ( abs ( x ) <= 12.7E+00 ) then

    y = 0.5E+00 * x**2

    q = exp ( - y ) * b0 / ( abs ( x ) - b1 &
      + b2 / ( abs ( x ) + b3 &
      + b4 / ( abs ( x ) - b5 &
      + b6 / ( abs ( x ) + b7 &
      - b8 / ( abs ( x ) + b9 &
      + b10 / ( abs ( x ) + b11 ) ) ) ) ) )
!
!  12.7 < |X|
!
  else

    q = 0.0E+00

  end if
!
!  Take account of negative X.
!
  if ( x < 0.0E+00 ) then
    cdf = q
  else
    cdf = 1.0E+00 - q
  end if

  return
end subroutine normal_01_cdf

subroutine normal_cdf ( x, a, b, cdf )
!
!*******************************************************************************
!
!! NORMAL_CDF evaluates the Normal CDF.
!
!
!  Discussion:
!
!    The Normal CDF is related to the Error Function ERF(X) by:
!
!      ERF ( X ) = 2 * NORMAL_CDF ( SQRT ( 2 ) * X ) - 1.0.
!
!  Modified:
!
!    23 February 1999
!
!  Author:
!
!    John Burkardt
!
!  Parameters:
!
!    Input, real X, the argument of the CDF.
!
!    Input, real A, B, the parameters of the PDF. (a=mean, b=standard devation)
!    0.0E+00 < B.
!
!    Output, real CDF, the value of the CDF.
!
  implicit none
!
  real(8), intent(in) :: a
  real(8), intent(in) :: b
  real(8), intent(out):: cdf
  real(8), intent(in) :: x
  real(8) :: y
!
  y = ( x - a ) / b

  call normal_01_cdf ( y, cdf )

  return
end subroutine normal_cdf

Subroutine rand_normal (mean,stdev,c)
implicit none
    Real(8), intent(in) :: mean,stdev
    real(8),intent(out) :: c
    real(8) :: temp(2),theta, Pi,r
    Pi= 3.1415926535897932
      IF(stdev <= 0.0d0) THEN

        WRITE(*,*) "Standard Deviation must be +ve"
      ELSE
        CALL RANDOM_NUMBER(temp)
        r=(-2.0d0*log(temp(1)))**0.5
        theta = 2.0d0*PI*temp(2)
        c= mean+stdev*r*sin(theta)
      END IF
end subroutine rand_normal    
    
end module 